#
#-##########################################################
#
# EDITS:
#
# 04/27/16 [htruax]: Changed read_doc_boxes( ) to look for '# ***** Calculations *****' as an indicator to stop
# reading '# ***** Field Data *****' boxes. The '# ***** Field Data *****' section has been moved from just before the
# Profiles to just after them.
#
# 12/17/18 [htruax]: Quick tweak to handle a startup from Visual Studio Code for a .rbw code file. Did not
# track down all comments and adjust them. This is a temporary version of this tool anyways, will soon have a
# 'C#' version.
#
#-##########################################################
#

#
#-#####################################################################
#
# This is Version 3 of the Blankety-Blank Application. We can now read from and write to the clipboard as well as starting
# the application from the SciTE Tools menu even while another instance of this application is currently running, so you can
# have two [usually] instances running, 1 from SciTE which closes when the 'Copy To Clipboard' button is selected [or the Exit
# menu button is selected, or the 'X' window button is clicked on] and another started previously from the Calc Tools folder.
#
#-#####################################
#
# Starting From the SciTE Tools Menu:
#
#-#####################################
#
# To use this new feature you will need to make some adjustments to the SciTEGlobal.properties file [SciTE Options Tab] in
# the section where we have our custom tools - near the bottom. Here is an example using command number 6 [make sure to
# remove the leading '# ' when copying this]:
#
# command.name.6.*=Blankety Blank
# command.6.*=C:\ruby\bin\rubyw.exe "\\PDC\FormsProgramming\Shared Code\Calc Tools\tk_blankety_blank_ver3.rbw" "$(FilePath)" "$(SelectionStartColumn)" "$(SelectionEndColumn)" "$(SelectionStartLine)" "$(SelectionEndLine)"
# command.save.before.6.*=1
# command.subsystem.6.*=0
# command.quiet.6.*=0
#
# Steps to follow:
#
# 1) From SciTE, you select [highlight] the code you wish to blank/reverse blank.
#
# 2) Either mouse click on the SciTE Tools menu to start this application or use the shortcut keys [i.e. 'Ctrl + 6' using above
#     example].
#	a) Application will start and will be fed the text that you have selected to either blank or reverse blank and will
#	   insert the selected text into the input region.
#	b) The Code Box List for the selected text will be populated using boxes found in the selected text.
#	c) The application reads the Ruby file for the document you started the application from and builds the Document Box
#	    List.
#	d) The application window title will be redisplayed with additional text " - STARTED FROM SciTE - " along with the
#	    document id number.
#
# 3) Do the desired blanking or reverse blanking.
#
# 4) Click on the 'Copy To Clipboard' button.
#	a) The clipboard is cleared and the Output Code [the blanked/reversed blanked code] is copied into it.
#	b) The application will exit, that is, it will stop executing and will NOT be minimized to the system tray.
#
# 5) In SciTE, you then paste the clipboard contents over the text you previously selected by pressing 'Ctrl + v'.
#
# Things to note:
#
#	a) This application automatically saves your SciTE file BEFORE the application runs.
#	b) The application must be eventually closed in order to use the SciTE function Key 'F5' - ONLY if you do not click
#	    on the 'Copy To Clipboard' button.
#
#-#####################################
#
# Using as a stand alone application:
#
#-#####################################
#
# This application DOES NOT need to be run from the SciTE Tools menu, it will still function as a stand alone application.
#
# Steps to follow:
#
# 1) From SciTE, you select [highlight] the code you wish to blank/reverse blank.
#
# 2) Copy the selected code to the clipboard via 'Ctrl + c'.
#
# 3) Either start the application or bring it up from the system tray.
#
# 4) Click on the 'Copy From Clipboard' button.
#	a) The selected SciTE text is inserted into the cleared input region of the tool.
#	b) The Code Box List for the selected text will be cleared and re-populated with boxes found in the selected text.
#	c) The Code Exclusion Boxes list will be cleared.
#
# 5) Do the desired blanking or reverse blanking.
#
# 6) Click on the 'Copy To Clipboard' button.
#	a) The clipboard is cleared and the Output Code [the blanked/reversed blanked code] is copied into it.
#
# 7) Either close or minimze the application.
#
# 8) In SciTE, you then paste the clipboard contents over the text you previously selected by pressing 'Ctrl + v'.
#
#-#####################################################################
#

require 'tk'

#
#-#####################################################################
#
# For displaying error messages.
#
require 'Win32API'

#
# Button styles:
#
$MB_OK = 0
$MB_OKCANCEL = 1
$MB_ABORTRETRYIGNORE = 2
$MB_YESNOCANCEL = 3
$MB_YESNO = 4
$MB_RETRYCANCEL = 5

#
# Icon styles:
#
$MB_ICONHAND = 16
$MB_ICONQUESTION = 32
$MB_ICONEXCLAMATION = 48
$MB_ICONASTERISK = 64

#
# Default button:
#
$MB_DEFBUTTON1 = 0
$MB_DEFBUTTON2 = 256
$MB_DEFBUTTON3 = 512

#
# Return values:
#
$IDOK = 1
$IDCANCEL = 2
$IDABORT = 3
$IDRETRY = 4
$IDIGNORE = 5
$IDYES = 6
$IDNO = 7

def messageBox(message, title = 'Message', style = 0)
	mb = Win32API.new("user32", "MessageBox", ['L', 'P', 'P', 'L'], 'L')
	return mb.Call(0, message, title, style)
end

#
#-#####################################################################
#
# Window drawing section.
#
root_title = "BLANKETY BLANK 3.1"

root = TkRoot.new() { title root_title }

root.geometry('+29+80')

#
#-#####################################################################
#
# List Frame: Frame for the Code boxes to be blanked and the Document boxes.
#
list_frame = TkFrame.new(root).pack(:side=>"top", :pady=>5)

#
#-#####################################################################
#
# List Frame: Code Frame - Boxes from the supplied Ruby Code.
#
code_frame = TkFrame.new(list_frame, :borderwidth=>5, :relief=>:ridge).pack(:side=>"left", :fill=>"x", :padx=>5)

code_label = TkLabel.new(code_frame, :text=>"Ruby Code: Single / Multi Blank Capability", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

#
#-#################################################
#
# Code Frame: Code Boxes
#
code_box_list_frame = TkFrame.new(code_frame, :borderwidth=>3, :relief=>:ridge).pack("side"=>"left", "pady" =>5, "padx"=>5)

code_box_list_label = TkLabel.new(code_box_list_frame, :text=>"Box List", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

code_box_sel_butt_frame = TkFrame.new(code_box_list_frame).pack("side"=>"bottom")

code_box_sel_mode_button = TkButton.new(code_box_sel_butt_frame, :text=>"Select Single", :background=>"lightgrey", :foreground=>"blue", :width=>11, "font"=>"arial 10").pack(:pady=> 5)

code_box_sel_del_frame = TkFrame.new(code_box_sel_butt_frame).pack(:pady=> 5)

code_box_clear_button = TkButton.new(code_box_sel_del_frame, :text=>"Delete List", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"left", :padx=> 5)

code_box_clear_mode_button = TkButton.new(code_box_sel_del_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"right", :padx=>4)

code_box_list_bar_vertical = TkScrollbar.new(code_box_list_frame).pack("side"=>"right", "fill"=>"y")

code_box_list = TkListbox.new(code_box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>15, "height"=>8, :selectmode=>"extended", "font"=>"arial 10").pack(:side=>"left")

code_box_list.yscrollbar(code_box_list_bar_vertical)

#
#-#################################################
#
# Code Frame: Exclusion Boxes
#
code_excl_box_list_frame = TkFrame.new(code_frame, :borderwidth=>3, :relief=>:ridge).pack(:anchor=>"n", :side=>"left", :pady=>5, :padx=>4)

code_excl_button = TkButton.new(code_excl_box_list_frame, :text=>"Exclusion Boxes", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:pady=>5)

code_excl_del_frame = TkFrame.new(code_excl_box_list_frame).pack(:side=>"bottom")

code_excl_clear_sel_button = TkButton.new(code_excl_del_frame, :text=>"Delete Selection", :background=>"lightgrey", :foreground=>"blue", :width=>12, "font"=>"arial 10").pack(:pady=>5)

code_excl_clear_button = TkButton.new(code_excl_del_frame, :text=>"Delete List", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left", :padx=>5, :pady=>5)

code_excl_clear_mode_button = TkButton.new(code_excl_del_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"right", :padx=>4, :pady=>5)

code_excl_box_list_bar_vert = TkScrollbar.new(code_excl_box_list_frame).pack("side"=>"right", "fill"=>"y")

code_excl_box_list = TkListbox.new(code_excl_box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>15, "height"=>7, :selectmode=>"extended", "font"=>"arial 10").pack(:side=>"left")

code_excl_box_list.yscrollbar(code_excl_box_list_bar_vert)

#
#-#####################################################################
#
# List Frame: Document Frame - All boxes from the document.
#
doc_frame = TkFrame.new(list_frame, :borderwidth=>5, :relief=>:ridge).pack(:side=>"left", :fill=>"x", :padx=>5)

doc_label = TkLabel.new(doc_frame, :text=>"Document: Single Blank Only", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

#
#-#################################################
#
# Document Frame: Document Boxes
#
doc_box_list_frame = TkFrame.new(doc_frame, :borderwidth=>3, :relief=>:ridge).pack("side"=>"left", "pady" =>5, "padx"=>5)

doc_box_list_label = TkLabel.new(doc_box_list_frame, :text=>"Box List", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

doc_box_sel_butt_frame = TkFrame.new(doc_box_list_frame).pack("side"=>"bottom")

doc_box_sel_mode_button = TkButton.new(doc_box_sel_butt_frame, :text=>"Select Single", :background=>"lightgrey", :foreground=>"blue", :width=>11, "font"=>"arial 10").pack(:pady=> 5)

doc_box_sel_del_frame = TkFrame.new(doc_box_sel_butt_frame).pack(:pady=> 5)

doc_box_clear_button = TkButton.new(doc_box_sel_del_frame, :text=>"Delete List", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"left", :padx=> 5)

doc_box_clear_mode_button = TkButton.new(doc_box_sel_del_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"right", :padx=>4)

doc_box_list_bar_vertical = TkScrollbar.new(doc_box_list_frame).pack("side"=>"right", "fill"=>"y")

doc_box_list = TkListbox.new(doc_box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>15, "height"=>8, :selectmode=>"extended", "font"=>"arial 10").pack(:side=>"left")

doc_box_list.yscrollbar(doc_box_list_bar_vertical)

#
#-#################################################
#
# Document Frame: Inclusion Boxes
#
doc_incl_box_list_frame = TkFrame.new(doc_frame, :borderwidth=>3, :relief=>:ridge).pack(:anchor=>"n", :side=>"left", :pady=>5, :padx=>4)

doc_incl_button = TkButton.new(doc_incl_box_list_frame, :text=>"Inclusion Boxes", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:pady=>5)

doc_incl_del_frame = TkFrame.new(doc_incl_box_list_frame).pack(:side=>"bottom")

doc_incl_clear_sel_button = TkButton.new(doc_incl_del_frame, :text=>"Delete Selection", :background=>"lightgrey", :foreground=>"blue", :width=>12, "font"=>"arial 10").pack(:pady=>5)

doc_incl_clear_button = TkButton.new(doc_incl_del_frame, :text=>"Delete List", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left", :padx=>5, :pady=>5)

doc_incl_clear_mode_button = TkButton.new(doc_incl_del_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"right", :padx=>4, :pady=>5)

doc_incl_box_list_bar_vert = TkScrollbar.new(doc_incl_box_list_frame).pack("side"=>"right", "fill"=>"y")

doc_incl_box_list = TkListbox.new(doc_incl_box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>15, "height"=>7, :selectmode=>"extended", "font"=>"arial 10").pack(:side=>"left")

doc_incl_box_list.yscrollbar(doc_incl_box_list_bar_vert)

#
#-#####################################################################
#
# This frame contains the Multi Blank group size and Custom Blanking areas.
#
misc_frame = TkFrame.new(list_frame, :borderwidth=>5, :relief=>:ridge).pack(:side=>"left", :fill=>"x", :padx=>5)

#
#-#################################################
#
# Custom Blanking:
#
custom_frame = TkFrame.new(misc_frame, :borderwidth=>3, :relief=>:ridge).pack(:side=>"top", :fill=>"x", :pady=>5, :padx=>5)

custom_label = TkLabel.new(custom_frame, :text=>"Custom BLANKETY BLANK Style - Defaults to: if ( )", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>"x")

custom_button_frame = TkFrame.new(custom_frame).pack(:side=>"bottom", :pady=>5, :padx=>5)

cust_clear_button = TkButton.new(custom_button_frame, :text=>"Delete Custom", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left", :padx=>5)

cust_clear_mode_button = TkButton.new(custom_button_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"left", :padx=>5)

custom_bar_horiz = TkScrollbar.new(custom_frame, "orient"=>"hor").pack("side"=>"bottom", "fill"=>"x")

custom_bar_text = TkText.new(custom_frame, "wrap"=>"none", "width"=>45, "height"=>1, "font"=>"arial 10").pack("fill"=>"x", "expand"=>true)
custom_bar_text.xscrollbar(custom_bar_horiz)

#
#-#################################################
#
# Multi Blanking:
#
multi_frame = TkFrame.new(misc_frame, :borderwidth=>3, :relief=>:ridge).pack(:side=>"bottom", :fill=>:x, :pady=>5, :padx=>5)

multi_label = TkLabel.new(multi_frame, :text=>"Multi Blank", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>"x")

group_size_frame = TkFrame.new(multi_frame, :borderwidth=>1, :relief=>:ridge).pack(:side=>:left, :pady=>5, :padx=>10)

group_label = TkLabel.new(group_size_frame, :text=>"Group Size", :width=>10, "font"=>"arial 10").pack(:side=>:left)

multi_boxes_group_num = TkVariable.new

multi_boxes_group_num_entry = TkEntry.new(group_size_frame, :textvariable=>multi_boxes_group_num, "width"=>3, "font"=>"arial 10").pack(:side=>:left)

multi_button_frame = TkFrame.new(multi_frame).pack(:side=>:left, :pady=>5, :padx=>5)

multi_clear_button = TkButton.new(multi_button_frame, :text=>"Delete Size", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left")

multi_clear_mode_button = TkButton.new(multi_button_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"left", :padx=>10)

#
#-#####################################################################
#
# Main buttons:
#
# Input Code Area:
#
blank_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"x")

copy_from_clipboard_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "padx"=>10, "pady"=>5)
copy_from_clipboard_button = TkButton.new(copy_from_clipboard_button_frame, "width"=>16, :background=>"lightgrey", :text=>"Copy From Clipboard", :foreground=>"blue", "font"=>"arial 10").pack()

s_blank_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5)
s_blank_button = TkButton.new(s_blank_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Single Blank", :foreground=>"blue", "font"=>"arial 10").pack()

m_blank_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "padx"=>10, "pady"=>5)
m_blank_button = TkButton.new(m_blank_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Multi Blank", :foreground=>"blue", "font"=>"arial 10").pack()

c_blank_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5)
c_blank_button = TkButton.new(c_blank_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Custom Blank", :foreground=>"blue", "font"=>"arial 10").pack()

r_blank_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "padx"=>10, "pady"=>5)
r_blank_button = TkButton.new(r_blank_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Reverse Blank", :foreground=>"blue", "font"=>"arial 10").pack()

clear_input_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5)
clear_input_button = TkButton.new(clear_input_button_frame, "width"=>12, :background=>"lightgrey", :text=> "Delete Code", :foreground=>"blue", "font"=>"arial 10").pack()

get_boxes_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "padx"=>10, "pady"=>5)
get_boxes_button = TkButton.new(get_boxes_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Code Box List", :foreground=>"blue", "font"=>"arial 10").pack()

delete_box_list_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5)
delete_box_list_button = TkButton.new(delete_box_list_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Delete All Lists", :foreground=>"blue", "font"=>"arial 10").pack()

input_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"both", "expand"=>true)

in_text_frame = TkFrame.new(input_frame).pack("side"=>"left", "fill"=>"both", "padx"=>10, "expand"=>true)
in_text_bar_vert = TkScrollbar.new(in_text_frame).pack("side"=>"right", "fill"=>"y")
in_text = TkText.new(in_text_frame, "width"=>113, "height"=>12, "font"=>"arial 10").pack("side"=>"bottom", "fill"=>"both", "expand"=>true)

in_text.yscrollbar(in_text_bar_vert)

in_text.focus

#
# Output Code Area:
#
output_button_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"both", "expand"=>true)

copy_to_clipboard_button_frame = TkFrame.new(output_button_frame).pack("side"=>"left", "pady"=>5, "padx"=>10)
copy_to_clipboard_button = TkButton.new(copy_to_clipboard_button_frame, "width"=>16, :background=>"lightgrey", :text=>"Copy To Clipboard", :foreground=>"blue", "font"=>"arial 10").pack()

output_label_frame = TkFrame.new(output_button_frame).pack("side"=>"left", "pady"=>5)
output_label = TkLabel.new(output_label_frame, :text=>"Output Code - NOTE: ALL code is selected by default when copying to Clipboard if no code has been selected.", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 12").pack()

output_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"both", "expand"=>true)

out_text_frame = TkFrame.new(output_frame).pack("side"=>"left", "padx" => 10, "fill"=>"both", "expand"=>true)
out_text_bar_vert = TkScrollbar.new(out_text_frame).pack("side"=>"right", "fill"=>"y")
out_text = TkText.new(out_text_frame, "width"=>113, "height"=>12, "font"=>"arial 10").pack("side"=>"bottom", "fill"=>"both", "expand"=>true)

out_text.yscrollbar(out_text_bar_vert)

bottom_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"x", "pady" => 5)

#
# ###############################################################
#
# Globals:
#
cur_dir = "C:/temp/work"

box_array = Array.new

#
# This tells us if this app was started from a code editor. Set to 'true' in the if (ARGV[0])..end expression
# further down.
#
fromCodeEditor = false

#
# ###############################################################
#
# Methods:
#
def process_errors(error, var, type)

	e_type = (type.eql?("W") ? "Warning" : "Error")

	err_msg = [
	"  #{e_type.upcase} - Duplicate Box: #{var}\t\n\n  This box is already in the list.\n",
	"  #{e_type.upcase} - Duplicate Box: #{var}\t\n\n  This box is in the Exclusion Boxes list.\n",
	"  #{e_type.upcase} - Duplicate Box: #{var}\t\n\n  This box is in the Inclusion Boxes list.\n",
	"  #{e_type.upcase} - #{var} Cannot be a trigger box - This is a calculated box.\n",
	"  Application was started from a code editor buffer that did not contain\n  the Ruby code from a document.\n",
	"  Application was started from a code editor buffer with the wrong number of arguments.\n",
	"  #{e_type.upcase} - process_sub_string( ) has produced a nil string. Application will Abort.\n",
	"  #{e_type.upcase} - Application could not locate the C:\\ruby\\scite\\SciTEGlobal.properties file which is needed to read the tab size value.\n\n You can not invoke this application from SciTE, use instead as a stand alone application where you can first copy your selection to the Clipboard.",
	"  #{e_type.upcase} - Tab size value from C:\\ruby\\scite\\SciTEGlobal.properties #{var} -  Application will default to a tab size value of 8.\n\n If this default tab size does not retrieve the selected text correctly use this application as a stand alone tool where you can copy your selection to the Clipboard.",
	"  #{e_type.upcase} - Application could not locate the tab size value from C:\\ruby\\scite\\SciTEGlobal.properties.\n\n You can not invoke this application from SciTE, use instead as a stand alone application where you first copy your selection to the Clipboard.",
	"  Application was started from Visual Studio Code with incorrect second argument.\n\nExpecting \"RC\"."
	]

	#
	# Using bitwise operations, error is set on a bit by bit basis, where each bit represents an error message.
	#
	if (error > 0)

		mask = 1

		#
		# Math.frexp(error) returns a two element array containing the normalized fraction (a Float) and exponent
		# (a Fixnum) of error. In our case, the fraction is 0.5 and the exponent is a power of 2, as our error is 1 x 2
		# raised to 0,1,2,3,... So this returns 0.5 x 2[a power]. So, say error = 4, which is 1 x 2[2], which is the 3rd
		# error in our list. .frexp returns [0.5, 3] - we want the 3 [3rd error], we use 3 - 1 since we are starting
		# from 0 not 1. Just doing this so that I do not have to keep updating the 0.upto( ) each time I add a new error.
		#
		0.upto(Math.frexp(error)[1]-1) do |shift|
			if (((mask << shift) & error) > 0)
				messageBox(err_msg[shift], "#{e_type}!", $MB_ICONEXCLAMATION|$MB_OK)
			end
		end
	end
end

def get_file_name(dir, type)

	ftypes = case (type)
			when ("RB"): [ ["RB files", '.rb'] ]
			when ("RBW"): [ ["RBW files", '.rbw'] ]
		end

	Tk.getOpenFile(:filetypes=>ftypes, :initialdir=>dir)
end

def read_doc_boxes(source)

	box_array = Array.new
	tmp_array = Array.new

	File.open(source, "r") do |src_file|

		#while ((line = src_file.gets) and (line !~ /\*+\s*Profile Classes\s*\*+/))
		while ((line = src_file.gets) and (line !~ /\*+\s*Calculations\s*\*+/))
			box_array.push($1) if (line =~ /(box\d+_\d+)/)
		end
	end

	#
	# Sort the array:
	#
	box_array.each { |box|
		box.scan(/box(\d+)_(\d+)/) { tmp_array.push([$1.to_i, $2.to_i]) }
	}

	tmp_array.sort!
	box_array.clear

	tmp_array.each_index { |index|
		box_array.push("box#{tmp_array[index][0]}_#{tmp_array[index][1]}")
	}

	box_array
end

#
# This will reverse the blanked code to the original form. 05/13/14 [HET]: Keeping this method pretty much the same regarding
# looking for the various blanking patterns to reverse [April/Bob/Harry] as these have been in use for several years.
#
def reverse_blank_button(out_text, in_text)

	box_array = Array.new
	code_array = Array.new
	input_array = Array.new
	temp_array = Array.new
	print_array = Array.new
	blank_array = Array.new
	comment_array = Array.new

	output_str = String.new

	#
	# input_array is set via splitting the supplied code at the newline character, each array element is a line of data.
	#
	input_array = in_text.get("1.0", "end").split(/\n/)

	return if (input_array.length == 0)

	#
	# Process the input_array by looking for a blankety-blank pattern, or patterns in the case of a multi-blank situation.
	#
	process_code = 0
	reverse_code = 0
	end_cnt = 0

	process_print = false
	process_blank = false
	check_for_end_exp = false

	pattern_string = output_str = ""

	input_array.each { |code_line|
#~ out_text.insert("end", "At top - code_line = #{code_line}\n")
#~ out_text.insert("end", "At top - pattern_string = #{pattern_string}\n")

		#
		# Look for the blankety-blank pattern(s) - April/Bob/Harry/Custom.
		#
		# An 'if' may or may not be the start of a blankety blank pattern, we start by assuming it is, but we also keep
		# a copy of each code line in case it turns out not to be a blankety blank pattern, in which case we will turn off the
		# pattern recognition flag and output the previously copied code as it was inputted.
		#
		# For blankety blank patterns in process:
		#
		# We will skip blank lines (new line char is stripped, so a blank line will have 0 length in the input_array), we will
		# oversee the line spacing when we output the reversed code.
		#
		if (!pattern_string.empty?)
			temp_array.push(code_line)
			next if (code_line.strip.length == 0)
		end

		case code_line.strip

			when (/\A\s*if(\s+|\()/)

				#
				# NOTE: An 'if' statement - could have nothing between the parens (empty custom blanking).
				# ALSO: This could be the start of a multi-blankety blank component, or a stand-alone blankety blank
				# pattern.
				#
				if (pattern_string.empty?)

					temp_array.push(code_line)
					pattern_string = "if"

				elsif ((pattern_string == "if blank else") or (pattern_string == "if"))

					#
					# This 'if' is in the code section, it is not a blankety blank 'if'.
					#
					end_cnt += 1

					#
					# First get any comments since we saw the 'initial if'.
					#
					if (pattern_string == "if")
						comment_array.each { |comment|
							code_array.push(comment)
						}
					end

					code_array.push(code_line.sub(/\t/,''))
					pattern_string += " code"

				elsif ((pattern_string == "if blank else code") or (pattern_string == "if code"))
					end_cnt += 1
					code_array.push(code_line.sub(/\t/,''))
				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

			when (/\A\s*else\z/)

				#
				# We need to be careful here, as an 'else' could go with a code section 'if', or be the 'else'
				# that goes with the blankety blank format.
				#
				if (pattern_string == "if blank")
					pattern_string += " else" 
				elsif (pattern_string == "if code")
					if (end_cnt > 0)
						code_array.push(code_line.sub(/\t/,''))
					else
						pattern_string += " else" 
					end
				elsif (pattern_string == "if blank else code")
					if (end_cnt > 0)
						code_array.push(code_line.sub(/\t/,''))
					else
						#
						# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
						#
						process_code = 1
					end
				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

			when (/\A\s*end\z/)

				#
				# We need to be careful here, as an 'end' could go with a code section 'if', or be the 'end'
				# that goes with the blankety blank format, or is an 'end' to a 'case', 'while', 'until', 'for',
				# 'unless', or 'loop do' expression (yikes !).
				#
				if (pattern_string == "if code else blank")
					pattern_string += " end" 
				elsif (pattern_string == "if blank else code")
					if (end_cnt > 0)
						code_array.push(code_line.sub(/\t/,''))
						end_cnt -= 1
					else
						pattern_string += " end" 
					end
				elsif (pattern_string == "if code")
					if (end_cnt > 0)
						code_array.push(code_line.sub(/\t/,''))
						end_cnt -= 1
					else
						#
						# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
						#
						process_code = 1
					end
				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

			when (/(\A\s*print\s*')(box)(\d+)(_)(\d+)(',\s*box\d+_\d+\z)/)
#~ out_text.insert("end", "a print: $3 = #{$3}, $5 = #{$5}\n")
#~ out_text.insert("end", "a print: pattern_string = #{pattern_string}\n")

				#
				# The 'print' section comes at the very end of a properly formed blankety blank formation.
				# If this is a blankety blank print, check to see if this is the end of the pattern by comparing to the
				# blanked boxes array.
				#
				if ((pattern_string == "if blank else code end") or (pattern_string == "if code else blank end"))
					pattern_string += " print"
					process_print = true

				elsif ((pattern_string == "if blank else code end print") or (pattern_string == "if code else blank end print"))

					process_print = true

				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

				if (process_print)

					#
					# This print statement is a blankety blank print (so far), check to see if we are done with
					# the pattern.
					#
					print_array.push([$3.to_i,$5.to_i])
					print_array.sort!
#~ out_text.insert("end", "print_array.push([#{$3},#{$5}])\n")

					if (blank_array.eql?(print_array))

						#
						# Done with the current blankety blank pattern.
						#
						process_code = 2
					end

					process_print = false
				end

			when (/(\A\s*)(box)(\d+)(_)(\d+)(\s*=\s*''\z)/)

				#
				# A box is blanked - it is conceivable that this could be a valid code section component.
				# If the very first line in the blankety blank code section is a blank assignment to a box,
				# this will be construed as the start of the blankety blank blank section, so to get around
				# this 'mode switch' the user can make this line a comment and it will be properly treated
				# as a code section component.
				#
				if ((pattern_string == "if") or (pattern_string == "if code else"))
					pattern_string += " blank"
					process_blank = true
				elsif ((pattern_string == "if blank") or (pattern_string == "if code else blank"))
					process_blank = true
				elsif ((pattern_string == "if code") or (pattern_string == "if blank else code"))
					code_array.push(code_line.sub(/\t/,''))
				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

				if (process_blank)
					blank_array.push([$3.to_i,$5.to_i])
					blank_array.sort!
					process_blank = false
				end

			when (/\A\s*#/)

				#
				# A comment - ignore comments in blankety-blank blank and print sections.
				#
				case pattern_string

					when /\Aif\z/

						#
						# Remember the comment, use it if we move into a code section - we do not yet know
						# for sure.
						#
						comment_array.push(code_line.sub(/\t/,''))

					when /\Aif blank\z|\Aif code else\z|\Aif code else blank\z|\Aif blank else code end\z|\Aif blank else code end print\z|\Aif code else blank end\z|\Aif code else blank end print\z/

						#
						# Ignore the comment.
						#

					when /\Aif code\z|\Aif blank else\z|\Aif blank else code\z/

						#
						# Process the comment - switch to code mode if appropriate.
						#
						code_array.push(code_line.sub(/\t/,''))
						pattern_string += " code" if (pattern_string == "if blank else")

					else
						#
						# Not a blankety blank pattern - output the temp array and reset the pattern string
						# to empty.
						#
						process_code = 1
				end

			else

				#
				# Something other than an 'if/else/end/print/blank/comment' - defaults to a code section component
				# if we are processing a blankety blank component, else, we just output whatever it is and go to the
				# next index.
				#
				case pattern_string

					when /\Aif code\z|\Aif blank else code\z/

						code_array.push(code_line.sub(/\t/,''))

						#
						# We just inserted into the code_array, check if we inserted an expression that also
						# uses an 'end' keyword.
						#
						check_for_end_exp = true

					when /\Aif\z|\Aif blank else\z/

						#
						# First get any comments since we saw the 'initial if'.
						#
						if (pattern_string == "if")
							comment_array.each { |comment|
								code_array.push(comment)
							}
						end

						code_array.push(code_line.sub(/\t/,''))

						#
						# We just inserted into the code_array, check if we inserted an expression that also
						# uses an 'end' keyword.
						#
						check_for_end_exp = true

						pattern_string += " code"
#~ out_text.insert("end", "else:   code_line = #{code_line}\n")
#~ out_text.insert("end", "        pattern_string = #{pattern_string}\n")

					else
						#
						# Not a blankety blank pattern - output the temp array and reset the pattern
						# string to empty.
						#
						process_code = 1
				end

				if (check_for_end_exp)

					if (code_line =~ /\A\s*for\s+|\A\s*unless(\s+|\()|\A\s*case(\s+|\()|\A\s*while(\s+|\()|\A\s*until(\s+|\()|\A\s*loop\s+do/)
						end_cnt += 1
					end

					check_for_end_exp = false
				end
		end

		#
		# 03/12/15 [HET]: I switched to building the entire output code first via output_str then inserting that into the
		# out_text widget when I was trying to have SciTE automatically capture the output and overwrite the ruby source
		# file, but that did not work due to a bug in our version of SciTE. Thought it would be better to build the string first
		# then either insert it into the out_text widget or print it out to SciTE. Currently leaving the prior use of 
		# out_text.insert but commented them out. If we go back to the out_text.insert way we will first need to clear it at
		# the top of this routine like we used to do.
		#
		case process_code

			when 1

				#
				# Blankety blank pattern is not "complete", output the temp array if it contains anything, output the
				# current code_line value if we are not working on a pattern (pattern_string being empty, thus code_line
				# value will not be stored in temp_array).
				#
				temp_array.each { |temp_code|

					#~ out_text.insert("end", "#{temp_code}\n")
					output_str += "#{temp_code}\n"
				}

				#~ out_text.insert("end", "#{code_line}\n") if (pattern_string.empty?)
				output_str += "#{code_line}\n" if (pattern_string.empty?)

			when 2
#~ out_text.insert("end", "process_code = #{process_code}\n")

				#
				# We found a complete pattern - Now use the code array to produce the blankety blank reversal output.
				#
				end_cnt = 0
				prev_box = ''

				0.upto(code_array.length - 1) { |index|
			
					#~ out_text.insert("end", "#{code_array[index]}\n")
					output_str += "#{code_array[index]}\n"

					case code_array[index]

						when (/(\A\s*)(box)(\d+)(_)(\d+)(\s*(\+?|-?)=.*\z)/)

							#
							# Box assignment - build a list of boxes if we are inside an expression structure that terminates
							# with an 'end' keyword - for example: 'if'/'unless'/'case'/'while'/'until'/'for'/'loop do'.
							#
							box_array.push([$3.to_i,$5.to_i]) if ((!box_array.include?([$3.to_i,$5.to_i])) and (end_cnt > 0))

							this_box = $2+$3+$4+$5

							if (index == code_array.length - 1)

								#~ out_text.insert("end", "print '#{this_box}', #{this_box}\n")
								output_str += "print '#{this_box}', #{this_box}\n"

							elsif (code_array[index + 1] =~ /(\A\s*)(box\d+_\d+)(\s*(\+?|-?)=.*\z)/)
								if (end_cnt == 0)
									if (this_box != $2)
										#~ out_text.insert("end", "print '#{this_box}', #{this_box}\n\n")
										output_str += "print '#{this_box}', #{this_box}\n\n"
									end
								end
							else
								#~ out_text.insert("end", "print '#{this_box}', #{this_box}\n") if (end_cnt == 0)
								output_str += "print '#{this_box}', #{this_box}\n"  if (end_cnt == 0)
								#~ out_text.insert("end", "\n")
								output_str += "\n"
							end

						when (/\A\s*if(\s+|\()|\A\s*for\s+|\A\s*unless(\s+|\()|\A\s*case(\s+|\()|\A\s*while(\s+|\()|\A\s*until(\s+|\()|\A\s*loop\s+do/)

							#
							# An 'if'/'for'/etc.. statement - print boxes at end of outer 'if'/'for'/etc.. statement
							# (may be nesting). A list of all boxes that are assigned to will be built and then used to
							# produce the print statements once the outer 'end' statement is encountered (end_cnt == 0).
							#
							end_cnt += 1
							#~ out_text.insert("end", "\n")
							output_str += "\n"

						when (/\A\s*end\z/)

							#
							# Decrement the end_cnt and if equal to zero, output the print box statements.
							#
							end_cnt -= 1 if (end_cnt > 0)

							if (end_cnt == 0)

								#~ out_text.insert("end", "\n")
								output_str += "\n"

								box_array.sort!

								box_array.each_index { |i|

									box = "box#{box_array[i][0]}_#{box_array[i][1]}"
									#~ out_text.insert("end", "print '#{box}', #{box}\n")
									output_str += "print '#{box}', #{box}\n"
								}

								#
								# Empty the box_array.
								#
								box_array.clear
							end

							#~ out_text.insert("end", "\n") if (index != code_array.length - 1)
							output_str += "\n" if (index != code_array.length - 1)
						else
							#~ out_text.insert("end", "\n")
							output_str += "\n"
					end
				}
		end
#~ out_text.insert("end", "At bottom - process_code = #{process_code}\n")
#~ out_text.insert("end", "At bottom - pattern_string = #{pattern_string}\n")

		if (process_code > 0)

			#
			# Either we determined that the 'pattern' we were working on turned out not to be a blankety blank
			# pattern, or we just processed a blankety blank pattern. Init to recognise and process the next pattern.
			#
			pattern_string = ''
			temp_array.clear
			blank_array.clear
			print_array.clear
			code_array.clear
			comment_array.clear
			process_code = 0
		end
	}

	#
	# There could possibly be some leftover code in the temp_array.
	#
	temp_array.each { |temp_code|

		#~ out_text.insert("end", "#{temp_code}\n")
		output_str += "#{temp_code}\n"
	}

	# Clear the output area and insert the output_str:
	out_text.delete("1.0", "end")
	out_text.insert("end", output_str)

	# Set the focus to the output text window, and put the cursor at the first char on line one.
	out_text.focus
	out_text.mark_set('insert', '1.0')
end

def blank_button(mode, out_text, in_text, code_excl_box_list, doc_incl_box_list, custom_bar_text, group_num)

	blank_array = Array.new
	blank_page_num_array = Array.new
	print_array = Array.new
	code_array = Array.new
	incl_box_array = Array.new

	input_str = String.new

	box = String.new

	error_code = 0
	group_cnt = 1

	# Clear the output area, and check if we can proceed.
	out_text.delete("1.0", "end")

	# Just catching a user goof.
	group_num = 0 if (group_num < 0)

	# Multi-blanking and Custom blanking ignores the incl_box_array.
	incl_box_array = doc_incl_box_list.get(0, "end") if (mode == "single")

	input_str = in_text.get("1.0", "end").strip

	if (input_str.strip.length > 0)

		input_str.each { |code_line|

			case
				when (code_line =~ /(\s*print +')(box)(\d+)(_)(\d+)(', +box\d+_\d+)/)

					blank_page_num_array.push([$3.to_i,$5.to_i]) unless (blank_page_num_array.include?([$3.to_i, $5.to_i]))
				when (!code_line.strip.empty?): code_array.push(code_line)

				when (mode == "multiple")

					# We are multi-blanking, we need to determine if it is time to do it now, as blank lines are significant.
					if ((group_num == 0) or (group_num == group_cnt))

						# First produce a sorted blank_array and use it to make the print_array - should have done this
						# differently, but time is short and I am tired and it is 'Friday'.
						blank_page_num_array.sort!

						blank_page_num_array.each_index { |i|

							box = "box#{blank_page_num_array[i][0]}_#{blank_page_num_array[i][1]}"
							blank_array.push(box)
							print_array.push("print '#{box}', #{box}")
						}
	
						do_the_blank_thing(mode, out_text, code_excl_box_list, doc_incl_box_list, custom_bar_text, print_array, blank_array, code_array)

						# Init for next group.
						group_cnt = 1
						print_array.clear
						blank_array.clear
						blank_page_num_array.clear
						code_array.clear
						out_text.insert("end", "\n")
					else
						group_cnt += 1
					end
			end
		}

		# Check for any blanking that needs to be done, if we are single blanking, we will need to do it now.
		if ((print_array.length + blank_array.length + code_array.length) > 0)

			# First produce a sorted blank_array and use it to make the print_array - should have done this
			# differently, but time is short and I am tired and it is 'Friday'.
			blank_page_num_array.sort!

			blank_page_num_array.each_index { |index|

				box = "box#{blank_page_num_array[index][0]}_#{blank_page_num_array[index][1]}"
				blank_array.push(box)
				print_array.push("print '#{box}', #{box}")
			}

			do_the_blank_thing(mode, out_text, code_excl_box_list, doc_incl_box_list, custom_bar_text, print_array, blank_array, code_array)
		end

		# Set the focus to the output text window, and put the cursor at the first char on line one.
		out_text.focus
		out_text.mark_set('insert', '1.0')
	end
end

def do_the_blank_thing(mode, out_text, code_excl_box_list, doc_incl_box_list, custom_bar_text, print_array, blank_array, code_array)

	exclusion_box_array = Array.new
	inclusion_box_array = Array.new
	page_num_array = Array.new

	box = String.new
	box_str = String.new
	output_str = String.new

	# box_str will be the string of boxes for the if ( ) statement.
	box_str = ''

	if (mode == "custom")

		box_str = (custom_bar_text.strip.empty? ? "if ( )" : custom_bar_text)
	else
		#
		# Not doing custom blanking, so build the box_str for the if expression from the ruby code
		# and if not multi-blanking, add the inclusion boxes.
		#
		exclusion_box_array = code_excl_box_list.get(0, "end")

		code_array.each { |code_line|
			code_line.scan(/(box)(\d+)(_)(\d+)/) {
			
			if (!(blank_array.include?($1+$2+$3+$4) or exclusion_box_array.include?($1+$2+$3+$4) or page_num_array.include?([$2.to_i,$4.to_i])))
				page_num_array.push([$2.to_i,$4.to_i])
			end
			}
		}

		unless (mode == "multiple")

			inclusion_box_array = doc_incl_box_list.get(0, "end")

			inclusion_box_array.each { |box|

				box.scan(/(box)(\d+)(_)(\d+)/) {

					#
					# We have already made sure that the Exclusion and Inclusion lists do not intersect.
					#
					case
						when (blank_array.include?($1+$2+$3+$4)): process_errors(8, box, "W")
						when (page_num_array.include?([$2.to_i,$4.to_i])): # do nothing.
						else page_num_array.push([$2.to_i,$4.to_i])
					end
				}
			}
		end

		# Now sort the two-dimensional page_num_array, which will put the boxes in order correctly, note that this
		# is not a string sort, but an integer sort based on page number and box number for each box. A string sort
		# would say that 'box1_1' is greater than 'box10_1'.
		page_num_array.sort!

		# Now put the box parts back together and assign to box_str. box_str now holds all boxes that are not in
		# the exclusion array, not in the blank array, and are not duplicates - they came from the range of boxes
		# encapsulated in the code that is to be blanked, plus any boxes added per user request via the inclusion list
		# [non multi blank only].
		page_num_array.each_index { |i| 
			box_str += "box#{page_num_array[i][0]}_#{page_num_array[i][1]}"
		}

		box_str.gsub!(/(\d)(b)/, '\1 + \2')
	end

	# Now output the blanked code in the desired blanking style.
	output_str = ""
	output_str += (mode == "custom" ? "#{box_str}\n" : "if ((#{box_str}).strip.length > 0)\n\n")

	code_array.each { |print_line| output_str += "\t#{print_line}" }
	output_str += "else\n"
	blank_array.each { |print_line| output_str += "\t#{print_line} = ''\n" }
	output_str += "end\n\n"
	print_array.each { |print_line| output_str += "#{print_line}\n" }

	out_text.insert("end", output_str)
end

def process_doc_file(fname, doc_box_list_label, doc_box_list)

	doc_box_list_label.configure("text", "Box List #{File.basename(fname, ".*").upcase}")

	doc_box_list.delete(0, "end")

	tmp_array = read_doc_boxes(fname)

	tmp_array.each { |box| doc_box_list.insert("end", box) }

	# Select (show highlighting) first list element.
	doc_box_list.selection_set(0)
end

def process_input_code(code_box_list, in_text)

	page_num_array = Array.new
	box = String.new
	input_str = String.new

	# Clear the Box List.
	code_box_list.delete(0, code_box_list.index("end"))

	# Get input code.
	input_str = in_text.get("1.0", "end")

	input_str.each { |code_line|
		code_line.scan(/(box)(\d+)(_)(\d+)/) {
			page_num_array.push([$2.to_i,$4.to_i]) unless page_num_array.include?([$2.to_i,$4.to_i])
		}
	}

	# Now sort the two-dimensional page_num_array.
	page_num_array.sort!

	# Now put the box parts back together and insert into the Box List widget.
	page_num_array.each_index { |index|
		code_box_list.insert("end", "box#{page_num_array[index][0]}_#{page_num_array[index][1]}")
	}

	# Select (show highlighting) first list element.
	code_box_list.selection_set(0)
end

#
# Clears the in_text and out_text widgets, the associated Ruby Code 'Box List' and the 'Exclusion Boxes' list.
#
def clear_input_output_code_and_lists(in_text, out_text, code_box_list, code_excl_box_list)

	in_text.delete("1.0", "end")
	out_text.delete("1.0", "end")
	code_box_list.delete(0, "end")
	code_excl_box_list.delete(0, "end")
end

#
# This routine returns a sub-string from the current 'line' read from the file we are selecting from. We call this routine
# from process_current_selection( ), when we are processing either the first 'line' or the last 'line', or when the first
# and last lines are the same line.
#
# First Line Processing: We need to find the index position in 'line' that corresponds to the first_col as supplied from
# SciTE. Note that our index values start from 0 since we are indexing into a string [the supplied 'line' argument]. We
# need to have the tab size to do this. From our end of it, a tab is just a single character just like a space char or the
# letter 'A', but we use the tab size when we process a tab char so we can properly set our 'col' variable which in turn
# allows us to properly set our index values. Our 'col' value corresponds to SciTE column values which start from 1. 
#
# Last Line Processing: We need to find the index position in 'line' that corresponds to the last_col as supplied from
# SciTE. Very similar logic to First Line Processing. We return the sub-string starting from the start of the line and runs
# up to and includes the last index position.
#
# The 'mode' argument tells us if we are processing a sub-string that is on both the first and last line, in which case
# we set the first and last index values, and return the corresponding sub-string.
#
# We look at each byte in 'line', a value of 9 is the tab character.
#
def process_sub_string(mode, line, first_col, last_col, tab_size)

	sub_string = nil
	col = 1
	index = first_index = 0
	tab = tab_size + 1

	line.each_byte do |byte|

		(byte == 9 ? col = tab : col += 1)
		tab += tab_size unless (col < tab)
		index += 1

		case (mode)

			when ("FIRST")

				if (col == first_col)
					sub_string = line[index, line.length - index]
					break
				end

			when ("LAST")

				if (col == last_col)
					sub_string = line[0, index]
					break
				end

			when ("BOTH")

				first_index = index if (col == first_col)

				if (col == last_col)
					sub_string = line[first_index, index - first_index]
					break
				end
		end
	end

	unless (sub_string)

		process_errors(64, nil, "E")
		exit(0)
	end
	sub_string
end

#
# This routine is called when we startup from SciTE.
#
# This retrieves currently selected text from the ruby file. We can not just grab the text via "$(CurrentSelection)"
# [would be used in SciTE Global Properties file] because double quotes are somehow filtered out of the selection so you would 
# get #{box1_1} instead of "#{box1_1}", also, I saw mention somewhere [can not remember exactly where] that
# "$(CurrentSelection)" can only hold so many characters, I think 1000.
#
# Note that we need to retrieve the tab size value from SciTEGlobal.properties file [see the '# Indentation' section in that
# file] to properly interpret the tab chars ["\t"] in the selection we are processing so we grab the correct chars. It is assumed
# that users will not be using either the "Local Options" or the "User Options" files for setting the tab size. Testing has
# revealed that there is a defult tab size of 8 when the tabsize line in the SciTEGlobal.properties file has been commented
# out, or set to 0, regardless of using/not using the other items under the '# Indentation' section.
#
# Note that we only have the starting column of the first line and the stopping column of the last line, so we grab
# entire lines that fall between them. This is exactly what "$(CurrentSelection)" does. So users need to select text
# appropriately.
#
# Returns a string containing the selected Ruby code.
#
def process_current_selection(file, strt_col, end_col, strt_line, end_line, tab_size)

	ln_num = 1
	selection = ''
	process = 0
	same_line = (strt_line == end_line)

	File.open(file, "r") do |src_file|

		while (line = src_file.gets)

			if (ln_num == strt_line)

				if (same_line)

					case
						when ((strt_col == 1) and (end_col == 1)): selection = ''
						when (strt_col == 1)
							selection = process_sub_string("LAST", line, strt_col, end_col, tab_size)
						else
							selection = process_sub_string("BOTH", line, strt_col, end_col, tab_size)
					end
					process = 1 # 1 means we are done, break.
				else
					# Only interested in the starting column since the start and stop lines are not the same.
					selection = (strt_col > 1 ? process_sub_string("FIRST", line, strt_col, end_col, tab_size) : line[0, line.length])

					process = 2 # 2 means do not append to selection since we already did here.
				end
			elsif (ln_num == end_line)

				# We are only interested in the last column since we are on the last line.
				selection += (end_col > 1 ? process_sub_string("LAST", line, strt_col, end_col, tab_size) : '')
				process = 1 # 1 means we are done, break.
			end

			case (process)

				when (1): break
				when (2): process = 3
				when (3): selection += line
			end

			ln_num += 1
		end
	end
	selection
end

#
# Called from inside the startup from SciTE section once the existance of this file has been verified - see "if (ARGV[0])"
# block further down below.
#
# This routine reads the "SciTEGlobal.properties" file to retrieve the tab size. We look in the '# Indentation'
# area. We default to a tab size of 8 IF: The 'tabsize' line is commented out, or the value is set to zero - testing
# these cases has revealed that SciTE uses a tab size of 8. The user is given a Warning about this and told they
# can use this application as a stand alone version if the selected text is not as expected, i.e. the default size does
# not work.
#
# Returns:
#
# tab_size set to 'nil' if we could not determine the tab size, else we return a positive tab size value.
#
def get_tab_size(file)

	process = false
	tab_size = nil

	File.open(file, "r") do |src_file|

		while (line = src_file.gets)

			case (line.delete("\s").upcase)
			
				when (/^#INDENTATION$/): process = true

				when (/^#~?TABSIZE=/)

					if (process)

						process_errors(256, "is commented out", "W")
						tab_size = 8
						break
					end

				when (/^TABSIZE=(\d+)$/)

					if (process)

						if ($1.to_i > 0)

							tab_size = $1.to_i
						else
							process_errors(256, "is set to 0", "W")
							tab_size = 8
						end
						break
					end
					
				when (/^#FOLDING$/): break
			end
		end
	end
	tab_size
end

#
# ###############################################################
#
# Menu:
#
#		Read:
#			Read Ruby Code file to create form box list [displaying box numbers].
#
#		Exit:
#			Exit application.
#
main_menu = TkMenu.new(root)

read_menu = TkMenu.new(main_menu, :tearoff=>false)

#
# This reads the .RBW file and builds an array of Boxes defined in the document. 
#
# The form_box_list is then poulated with the box numbers. The Document Inclusion Boxes list is cleared.
#
read_menu.add('command', :label=> "RB (Ruby Code) File", :command=> proc {

	tmp_array = Array.new

	fname = get_file_name(cur_dir, "RB")

	unless (fname.empty?)

		cur_dir = File.dirname(fname)
		process_doc_file(fname, doc_box_list_label, doc_box_list)
		doc_incl_box_list.delete(0, "end")
	end
})

read_menu.add(:separator)

#
# This reads the .RBW file and builds an array of Boxes defined in the document. 
#
# The form_box_list is then poulated with the box numbers. The Document Inclusion Boxes list is cleared.
#
read_menu.add('command', :label=> "RBW (Ruby Code) File", :command=> proc {

	tmp_array = Array.new

	fname = get_file_name(cur_dir, "RBW")

	unless (fname.empty?)

		cur_dir = File.dirname(fname)
		process_doc_file(fname, doc_box_list_label, doc_box_list)
		doc_incl_box_list.delete(0, "end")
	end
})

main_menu.add('cascade', :menu=>read_menu, :label=>"Read")

main_menu.add('command', :label=> "Exit", :command=> proc { exit(0) })

#
# Adding main_menu to the root window.
#
root.menu(main_menu)

#
# ###############################################################
#
# Button bindings:
#
code_excl_button.bind("Button-1") {

	box_indices_array = Array.new
	exclusion_box_array = Array.new
	inclusion_box_array = Array.new
	tmp_array = Array.new

	exclusion_box_array = code_excl_box_list.get(0, "end")
	inclusion_box_array = doc_incl_box_list.get(0, "end")

	box_indices_array = code_box_list.curselection

	box_indices_array.each { |i|

		box = code_box_list.get(i)

		case
			when (exclusion_box_array.include?(box)): process_errors(1, box, "E")
			when (inclusion_box_array.include?(box)): process_errors(4, box, "E")
			else exclusion_box_array.push(box)
		end
	}

	exclusion_box_array.each { |box|
		box.scan(/(box)(\d+)(_)(\d+)/) { tmp_array.push([$2.to_i,$4.to_i]) }
	}

	tmp_array.sort!
	code_excl_box_list.delete(0, "end")

	tmp_array.each_index { |i|
		code_excl_box_list.insert("end", "box#{tmp_array[i][0]}_#{tmp_array[i][1]}")
	}
}

doc_incl_button.bind("Button-1") {

	box_indices_array = Array.new
	exclusion_box_array = Array.new
	inclusion_box_array = Array.new
	tmp_array = Array.new

	inclusion_box_array = doc_incl_box_list.get(0, "end")
	exclusion_box_array = code_excl_box_list.get(0, "end")

	box_indices_array = doc_box_list.curselection

	box_indices_array.each { |i|

		box = doc_box_list.get(i)

		case
			when (inclusion_box_array.include?(box)): process_errors(1, box, "E")
			when (exclusion_box_array.include?(box)): process_errors(2, box, "E")
			else inclusion_box_array.push(box)
		end
	}

	inclusion_box_array.each { |box|
		box.scan(/(box)(\d+)(_)(\d+)/) { tmp_array.push([$2.to_i,$4.to_i]) }
	}

	tmp_array.sort!
	doc_incl_box_list.delete(0, "end")

	tmp_array.each_index { |i|
		doc_incl_box_list.insert("end", "box#{tmp_array[i][0]}_#{tmp_array[i][1]}")
	}
}

code_box_sel_mode_button.bind("Button-1") {

	# Determine the current selectmode configuration, it will be either "extended", one or
	# more contiguous items selectable at a time, or "multiple", where more than one 
	# non-contiguous items selectable at a time. Switch configuration between the two: 
	# "browse" or "multiple".
	case code_box_list.cget("selectmode")

		when "extended"

			code_box_list.configure("selectmode", "multiple")
			code_box_sel_mode_button.configure("text", "Select Multiple")

		when "multiple"

			# User can select (highlight) multiple list elements at once.
			code_box_list.configure("selectmode", "extended")
			code_box_sel_mode_button.configure("text", "Select Single")

			# Clear all selected elements, then select (highlight) the last element selected,
			# which is the currently 'active' element - the last one clicked on.
			code_box_list.selection_clear(0, "end")
			code_box_list.selection_set(code_box_list.index("active"))
	end
}

doc_box_sel_mode_button.bind("Button-1") {

	# Determine the current selectmode configuration, it will be either "extended", one or
	# more contiguous items selectable at a time, or "multiple", where more than one 
	# non-contiguous items selectable at a time. Switch configuration between the two: 
	# "browse" or "multiple".
	case doc_box_list.cget("selectmode")

		when "extended"

			doc_box_list.configure("selectmode", "multiple")
			doc_box_sel_mode_button.configure("text", "Select Multiple")

		when "multiple"

			# User can select (highlight) multiple list elements at once.
			doc_box_list.configure("selectmode", "extended")
			doc_box_sel_mode_button.configure("text", "Select Single")

			# Clear all selected elements, then select (highlight) the last element selected,
			# which is the currently 'active' element - the last one clicked on.
			doc_box_list.selection_clear(0, "end")
			doc_box_list.selection_set(doc_box_list.index("active"))
	end
}

clear_input_button.bind("Button-1") {

	# Clear all objects associated with the input code, and clears the out_text widget.
	clear_input_output_code_and_lists(in_text, out_text, code_box_list, code_excl_box_list)

	# Set the focus to the input text window, and put the cursor at the first char on line one.
	in_text.focus
	in_text.mark_set('insert', '1.0')
}

copy_from_clipboard_button.bind("Button-1") {

	# Copies the contents from the clipboard, and does the following:
	#
	# 1) Clears the in_text and out_text widgets, the Ruby Code 'Box List', and the associated 'Exclusion Boxes' list.
	# 2) Inserts the clipboard contents into the in_text widget, and sets the cursor at the start of the text area.
	# 3) Builds the Ruby Code 'Box List' from the contents of the in_text widget.

	# 1)
	clear_input_output_code_and_lists(in_text, out_text, code_box_list, code_excl_box_list)

	# 2)
	in_text.insert("1.0", TkClipboard.get(nil))
	in_text.focus
	in_text.mark_set('insert', '1.0')

	# 3)
	process_input_code(code_box_list, in_text)
}

s_blank_button.bind("Button-1") {

	# Perform single mode blanking and if the OFF/ON toggle buttons are "ON" we will also delete the corresponding list(s).
	blank_button("single", out_text, in_text, code_excl_box_list, doc_incl_box_list, custom_bar_text.value, multi_boxes_group_num.value.to_i)

	code_box_list.delete(0, code_box_list.index('end')) if (code_box_clear_mode_button.cget("text") == "ON")
	code_excl_box_list.delete(0, code_excl_box_list.index('end')) if (code_excl_clear_mode_button.cget("text") == "ON")
	doc_box_list.delete(0, doc_box_list.index('end')) if (doc_box_clear_mode_button.cget("text") == "ON")
	doc_incl_box_list.delete(0, doc_incl_box_list.index('end')) if (doc_incl_clear_mode_button.cget("text") == "ON")
}

m_blank_button.bind("Button-1") {

	# Perform multiple mode blanking and if the OFF/ON toggle buttons are "ON" we will also delete the corresponding list(s).
	blank_button("multiple", out_text, in_text, code_excl_box_list, doc_incl_box_list, custom_bar_text.value, multi_boxes_group_num.value.to_i)

	code_box_list.delete(0, code_box_list.index('end')) if (code_box_clear_mode_button.cget("text") == "ON")
	code_excl_box_list.delete(0, code_excl_box_list.index('end')) if (code_excl_clear_mode_button.cget("text") == "ON")
	multi_boxes_group_num.value = "" if (multi_clear_mode_button.cget("text") == "ON")
}

c_blank_button.bind("Button-1") {

	# Perform custom mode blanking and if the OFF/ON toggle buttons are "ON" we will also delete the corresponding vars.
	blank_button("custom", out_text, in_text, code_excl_box_list, doc_incl_box_list, custom_bar_text.value, multi_boxes_group_num.value.to_i)

	custom_bar_text.value = "" if (cust_clear_mode_button.cget("text") == "ON")
}

r_blank_button.bind("Button-1") {
	reverse_blank_button(out_text, in_text)
}

get_boxes_button.bind("Button-1") { process_input_code(code_box_list, in_text) }

code_box_clear_button.bind("Button-1") { code_box_list.delete(0, code_box_list.index('end')) }

#
# This will delete selected boxes from the code_excl_box_list.
#
code_excl_clear_sel_button.bind("Button-1") {

	tmp_array = Array.new

	indices = code_excl_box_list.curselection

	if (indices.size > 0)

		0.upto(code_excl_box_list.size - 1) do |i|

			tmp_array.push(code_excl_box_list.get(i)) unless (indices.include?(i)) 
		end

		code_excl_box_list.delete(0, "end")

		tmp_array.each { |box| code_excl_box_list.insert("end", box) }
	end
}

#
# This will delete selected boxes from the doc_incl_box_list.
#
doc_incl_clear_sel_button.bind("Button-1") {

	tmp_array = Array.new

	indices = doc_incl_box_list.curselection

	if (indices.size > 0)

		0.upto(doc_incl_box_list.size - 1) do |i|

			tmp_array.push(doc_incl_box_list.get(i)) unless (indices.include?(i)) 
		end

		doc_incl_box_list.delete(0, "end")

		tmp_array.each { |box| doc_incl_box_list.insert("end", box) }
	end
}

code_excl_clear_button.bind("Button-1") { code_excl_box_list.delete(0, code_excl_box_list.index('end')) }

code_box_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the code_box_list when either the single or multi blank button is selected.
	#
	case code_box_clear_mode_button.cget("text")

		when "OFF" then code_box_clear_mode_button.configure("text", "ON")
		when "ON" then code_box_clear_mode_button.configure("text", "OFF")
	end
	}

code_excl_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the code_excl_box_list when either the single or multi blank button is selected.
	#
	case code_excl_clear_mode_button.cget("text")

		when "OFF" then code_excl_clear_mode_button.configure("text", "ON")
		when "ON" then code_excl_clear_mode_button.configure("text", "OFF")
	end
	}

doc_box_clear_button.bind("Button-1") { doc_box_list.delete(0, doc_box_list.index('end')) }

doc_incl_clear_button.bind("Button-1") { doc_incl_box_list.delete(0, doc_incl_box_list.index('end')) }

doc_box_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the doc_box_list when either the single or multi blank button is selected.
	#
	case doc_box_clear_mode_button.cget("text")

		when "OFF" then doc_box_clear_mode_button.configure("text", "ON")
		when "ON" then doc_box_clear_mode_button.configure("text", "OFF")
	end
	}

doc_incl_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the doc_incl_box_list when either the single or multi blank button is selected.
	#
	case doc_incl_clear_mode_button.cget("text")

		when "OFF" then doc_incl_clear_mode_button.configure("text", "ON")
		when "ON" then doc_incl_clear_mode_button.configure("text", "OFF")
	end
	}

cust_clear_button.bind("Button-1") { custom_bar_text.value = "" }

cust_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the custom_bar_text when either the single or multi blank button is selected.
	#
	case cust_clear_mode_button.cget("text")

		when "OFF" then cust_clear_mode_button.configure("text", "ON")
		when "ON" then cust_clear_mode_button.configure("text", "OFF")
	end
	}

multi_clear_button.bind("Button-1") { multi_boxes_group_num.value = "" }

multi_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the custom_bar_text when either the single or multi blank button is selected.
	#
	case multi_clear_mode_button.cget("text")

		when "OFF" then multi_clear_mode_button.configure("text", "ON")
		when "ON" then multi_clear_mode_button.configure("text", "OFF")
	end
	}

delete_box_list_button.bind("Button-1") {

	# Clears all box lists.
	code_box_list.delete(0, code_box_list.index('end'))
	code_excl_box_list.delete(0, code_excl_box_list.index('end'))
	doc_box_list.delete(0, doc_box_list.index('end'))
	doc_incl_box_list.delete(0, doc_incl_box_list.index('end'))
}

copy_to_clipboard_button.bind("Button-1") {

# debug - demonstrating how the "end-1c" and other variants work. "end" is just past the very last character of the line,
# it would be where the next insertion would take place. "end-1c" moves into the string from the end one character position.
#~ out_text.insert("end", "end-1c..end: #{in_text.get("end-1c","end")[0]}\n")
#~ out_text.insert("end", "end-2c..end-1c: #{in_text.get("end-2c","end-1c")[0]}\n")
#~ out_text.insert("end", "end-3c..end-2c: #{in_text.get("end-3c","end-2c")[0]}\n")

	#
	# Copies text from the out_text widget to the clipboard. If no text has been selected it defaults to
	# copying ALL text to the clipboard. If started from the SciTE Tools Menu the application will then exit(0).
	#
	# If no out_text text has been selected, we will do the following when copying all text:
	#
	# If the newline character is not on it's own line in the in_text widget [looking at that since that is what we used for setting
	# the out_text widget] we will copy all out_text minus the newline character since we do not want an extra line in our source
	# file when we Ctrl-V from the clipboard. If however, the newline is on it's own line in the in_text widget, then we do want the
	# newline character.
	#
	if (out_text.tag_ranges("sel").empty?)

		output = out_text.get("1.0", in_text.index("end-1c").split(".")[1].to_i > 0 ? "end - 2c" : "end - 1c")
	else
		output = out_text.get("sel.first", "sel.last")
	end

	#
	# TkClipboard.set first clears the clipboard.
	#
	TkClipboard.set(output.empty? ? '' : output, nil)

	exit(0) if (fromCodeEditor)
}

#
# ###############################################################
#
# This handles the tool startup from SciTE or Visual Studio Code.
#
# We first start by verifying if we have Ruby code that is associated with a document, else we display an error
# message and abort. We then verify if we have 5 arguments, else we display an error message and abort. The Ruby
# Code file is read to obtain the list of boxes in the document, and any selected code is automatically read and
# inserted into the code input area and the associated code box list is populated. We also display the document ID
# and a 'started from SciTE' message in the window title area.
#
if (ARGV[0])

	goodLaunch = false

	if (File.basename(ARGV[0]).downcase =~ /([a-z]{4})\d{4}\.(rb|rbw)/)

		#
		# Check for startup from Visual Studio Code:
		#
		#~ if (ARGV.length == 2)
		if (ARGV.length == 1)

			#~ if (ARGV[1] == "RC")

				# Read the Clipboard:
				in_text.insert("end", TkClipboard.get(nil))
				goodLaunch = true

				#
				# This is the whole thing - path and file name.
				#
				cur_file = ARGV[0]
			#~ else
				#~ process_errors(1024, nil, "E")
				#~ exit(0)
			#~ end

		#
		# If set up properly in the SciTE Global Options file, we should have 5 arguments - the file path/name,
		# the start column, the end column, the start line, and the stop line, in that order. Note that we do not
		# actually confirm proper order, just the number of arguments.
		#
		elsif (ARGV.length == 5)

			#
			# We need to extract the tabsize setting from C:\ruby\scite\SciTEGlobal.properties, down in the 
			# '# Indentation' section. Tab size is used in properly selecting the text in the ruby file for a 
			# document. If for some reason the file does not exist, or can not be found given the supplied path,
			# we display an error message and abort. If we find the file and either the tabsize field is commented
			# out or has a value of 0 we will give a warning message and default to a tabsize of 8 as revealed by
			# testing such a scenario.
			#
			if (File.exist?("C:/ruby/scite/SciTEGlobal.properties"))

				if (tab_size = get_tab_size("C:/ruby/scite/SciTEGlobal.properties"))
# debug:
#~ out_text.insert("end", "ARGV[1].to_i = #{ARGV[1].to_i}\nARGV[2].to_i = #{ARGV[2].to_i}\nARGV[3].to_i = #{ARGV[3].to_i}\nARGV[4].to_i = #{ARGV[4].to_i}\n")
# end of debug
					goodLaunch = true

					#
					# This is the whole thing - path and file name.
					#
					cur_file = ARGV[0]

					in_text.insert("end", process_current_selection(cur_file, ARGV[1].to_i, ARGV[2].to_i, ARGV[3].to_i, ARGV[4].to_i, tab_size))
				else
					process_errors(512, nil, "E")
					exit(0)
				end
			else
				process_errors(128, nil, "E")
				exit(0)
			end
		else
			process_errors(32, nil, "E")
			exit(0)
		end
	else
		process_errors(16, nil, "E")

		#
		# We shut the app down.
		#
		exit(0)
	end

	if (goodLaunch)

		#
		# Now flag that we started this tool from a code editor (SciTE or Visual Studio Code) so we will automatically
		# shut down when the 'Copy To Clipboard' button has been clicked.
		#
		fromCodeEditor = true

		root.configure("title", "#{root_title} - STARTED FROM A CODE EDITOR - #{cur_file}")

		process_doc_file(cur_file, doc_box_list_label, doc_box_list)
		process_input_code(code_box_list, in_text)
	end
end

#
# 03/11/15 [HET]: This no longer pertains, but decided to keep comments and commented code in place for possible future use.
#
# This calls exit_app( ) when the user clicks on the 'X' button to close the application window so we
# can prevent the application output from writing/overwriting [when selected text] the SciTE file.
#
# This is because this application can run as a stand alone app in which case we do not automatically write to the
# SciTE file OR it can run from the SciTE Tools menu as a command in which case the app will close upon blanking/reverse
# blanking or could close when the user Exits the app from the menu or exits from the 'X' button.
#
#~ TkRoot.new.protocol('WM_DELETE_WINDOW', proc{exit_app(fromCodeEditor)})

Tk.mainloop()
